"""Create the kernel factory for the Conv2d GW8 kernel."""

from dataclasses import dataclass
from itertools import product
from typing import Generator

from .. import generators as gen
from ..cuda.driver import DeviceAttributes

from ..util import divup, next_relative_prime
from .launch_params import LaunchParams
from .conv2d_gw8_params import Conv2dGw8Params
from .conv2d_stats import Conv2dStats
from .kernel_factory import KernelFactory, KernelSpec
from .kernel import get_full_kernel_name


@dataclass(frozen=True)
class Conv2dGw8Config:
    """Tile configuration for the Conv2d GW8 kernel."""

    groups: int = 8
    block_p: int = 16
    block_n: int = 1


def _get_configs(
    params: Conv2dGw8Params, _device_attr: DeviceAttributes, **_kwargs
) -> Generator[Conv2dGw8Config, None, None]:
    """Generate configurations for the Conv2d GW8 kernel."""
    # igrad is unused in this function
    max_groups = min(params.groups, 8)
    block_n_values = [block_n for block_n in [1, 2, 4] if block_n <= params.n]
    block_p_values = [
        block_p for block_p in [1, 2, 4, 8, 16, 32, 64] if block_p <= params.p
    ]
    if params.p not in block_p_values:
        block_p_values.append(params.p)
    groups_values = [groups for groups in [1, 2, 4, 8] if groups <= max_groups]
    if params.groups not in groups_values and params.groups <= max_groups:
        groups_values.append(params.groups)
    yield from (
        Conv2dGw8Config(groups=groups, block_p=block_p, block_n=block_n)
        for groups, block_p, block_n in product(
            groups_values, block_p_values, block_n_values
        )
    )


def _get_kernel_name(igrad=False) -> str:
    return "spio_conv2d_gw8_fprop" if not igrad else "spio_conv2d_gw8_dgrad"


def _get_kernel_spec(
    params: Conv2dGw8Params,
    config: Conv2dGw8Config,
    _device_attr: DeviceAttributes,
    igrad: bool = False,
) -> KernelSpec:
    """The code generator specs and launch parameters."""
    params.validate()

    r, s = params.r, params.s

    if igrad:
        n, c, h, w = params.n, params.c, params.p, params.q
        p, q = params.h, params.w
        padding_h, padding_w = (
            params.transpose_padding_h,
            params.transpose_padding_w,
        )
    else:
        n, c, h, w = params.n, params.c, params.h, params.w
        p, q = params.p, params.q
        padding_h, padding_w = params.padding_h, params.padding_w

    # Hardcoded parameter:
    group_width = params.group_width

    # Derived parameters
    c8 = c // 8
    groups = params.groups

    # Tiles
    block_n = min(config.block_n, n)
    block_p = min(config.block_p, p)
    block_q = 16 // block_n
    block_groups = min(config.groups, groups)

    # Derived Tiles
    block_c = block_groups * group_width
    block_c8 = block_c // 8
    block_w = block_q + s - 1
    blocks_n = divup(n, block_n)
    blocks_p = divup(p, block_p)
    blocks_q = divup(q, block_q)
    blocks_c = divup(c, block_c)
    blocks = blocks_n * blocks_p * blocks_q * blocks_c
    warps = block_groups
    threads = warps * 32

    launch_params = LaunchParams(grid=blocks, block=threads)

    kernel_name = _get_kernel_name(igrad=igrad)
    full_kernel_name = get_full_kernel_name(kernel_name, params)

    kernel_has_bias = params.has_bias and not igrad

    # With 16 bytes-per-element, smem effectively has 8 banks.
    num_smem_banks = 8

    smem_x_stride = next_relative_prime(block_n * block_c8, num_smem_banks)

    gen_specs = [
        gen.Macro({"SPIO_CONV_KERNEL": full_kernel_name}),
        gen.Fold("block_n", "n", block_n),
        gen.Fold("block_p", "p", block_p),
        gen.Fold("block_q", "q", block_q),
        gen.Fold("block_c", "c", block_c),
        gen.ParamsSpec(
            "Block",
            {
                "threads": threads,
            },
        ),
        gen.ParamsSpec(
            "Padding",
            {
                "h": padding_h,
                "w": padding_w,
            },
        ),
        gen.ParamsSpec("Mode", {"igrad": igrad, "has_bias": kernel_has_bias}),
        gen.Index(
            "BlockIdx",
            gen.Dims(
                block_n=blocks_n,
                block_p=blocks_p,
                block_q=blocks_q,
                block_c=blocks_c,
            ),
        ),
        gen.Index("InputIdx", gen.Dims(n=block_n, x=block_w, c8=block_c8)),
        gen.Tensor(
            "Input", gen.dtype.uint4, gen.Dims(n=n, y=h, x=w, c8=c8), constant=True
        ),
        gen.Tensor("Bias", gen.dtype.float2, gen.Dims(k8=c8, k2=4), constant=True),
        gen.Index("BiasIdx", gen.Dims(k8=block_c8, lane=32)),
        gen.Tensor("Output", gen.dtype.uint4, gen.Dims(n=n, p=p, q=q, k8=c8)),
        gen.Tensor("Weights", gen.dtype.uint4, gen.Dims(k=c, r=r, s=s), constant=True),
        gen.Tensor("SmemWeights", gen.dtype.uint4, gen.Dims(k=block_c, r=r, s=s)),
        gen.Tensor(
            "ConstSmemWeights",
            gen.dtype.uint4,
            gen.Dims(k8=block_c8, k=8, r=r, s=s),
            constant=True,
        ),
        gen.Index(
            "SmemWeightsLoadIdx",
            gen.Dims(k8=block_c8, repeat=4, k=8),
            dummies=["repeat"],
        ),
        gen.Tensor(
            "SmemInput",
            gen.dtype.uint4,
            gen.Dims(ping_pong=2, x=block_w, n=block_n, c8=block_c8),
            strides=gen.Strides(x=smem_x_stride),
        ),
        gen.Index(
            "SmemInputLoadIdx",
            gen.Dims(
                c8=block_c8,
                repeat=32 // (block_q * block_n),
                x=block_q,
                n=block_n,
            ),
            dummies=["repeat"],
        ),
        gen.Index("SmemOutputStoreIdx", gen.Dims(k8=block_c8, lane=32)),
        gen.Tensor(
            "SmemOutput",
            gen.dtype.half2,
            gen.Dims(q=block_q, n=block_n, k8=block_c8 + 1, k2=4),
        ),
        gen.Tensor(
            "ConstSmemOutput",
            gen.dtype.uint4,
            gen.Dims(q=block_q, n=block_n, k8=block_c8 + 1),
            constant=True,
        ),
        gen.Index("OutputStoreIdx", gen.Dims(n=block_n, q=block_q, k8=block_c8)),
        gen.Index("BlockQNIdx", gen.Dims(q=block_q, n=block_n)),
        gen.Fragment("Acc", gen.FragmentType.M16_N8_F32_C, "qn", "k"),
        gen.Fragment("In", gen.FragmentType.M16_K8_F16_A, "qn", "c"),
        gen.Fragment("Wgts", gen.FragmentType.N8_K8_F16_B, "c", "k"),
        gen.Tensor("WeightsReg", "Wgts", gen.Dims(r=r, s=s)),
        gen.Tensor("AccReg", "Acc", gen.Dims(p=r)),
    ]
    return KernelSpec(gen_specs=gen_specs, launch_params=launch_params)


conv2d_gw8_kernel_factory = KernelFactory(
    Conv2dGw8Params,
    Conv2dGw8Config,
    Conv2dStats,
    kernel_name=_get_kernel_name,
    configs=_get_configs,
    kernel_spec=_get_kernel_spec,
    kernel_source_file="conv2d_gw8.cu",
    src_module="spio.src",
    includes_module="spio.include",
    perf_model_skip_params=["group_width", "stride"],
)



"""CUDA code generators."""

from .generators import generate
from .gen_specs import GenSpecs
from .index import Index
from .tensor import Tensor
from .fragment_type import FragmentType
from .data_type import dtype
from .fragment import Fragment
from .fragment_index import FragmentIndex, FragmentLoadIndex
from .macros import Macro
from .params import ParamsSpec
from .checkerboard import Checkerboard
from .async_strip_loader import AsyncStripLoader
from .dim import Dim, dim_name_to_dim_or_fold_class_name
from .fold import Fold
from .dims import Dims, Strides
from .matmul import Matmul



"""Test the dtype enumeration"""

import spio.generators as gen


def test_dtype_sizes():
    """Test the sizes of the data types."""
    assert gen.dtype.float.value.size == 4
    assert gen.dtype.float2.value.size == 8
    assert gen.dtype.float4.value.size == 16
    assert gen.dtype.half.value.size == 2
    assert gen.dtype.half2.value.size == 4
    assert gen.dtype.unsigned.value.size == 4
    assert gen.dtype.uint2.value.size == 8
    assert gen.dtype.uint4.value.size == 16
    assert gen.dtype.int32.value.size == 4



"""A dataclass for defining launch parameters for CUDA kernels."""

from dataclasses import dataclass
from typing import Tuple, Union


@dataclass
class LaunchParams:
    """Launch parameters for CUDA kernels.

    Attributes:
        grid: The grid size.
        block: The block size.
        shared_mem_bytes: The amount of dynamic shared memory in bytes.
    """

    grid: Union[int, Tuple[int, ...]]
    block: Union[int, Tuple[int, ...]]
    shared_mem_bytes: int = 0

    def __post_init__(self):
        assert self.grid > 0
        assert self.block > 0
        if isinstance(self.grid, int):
            self.grid = (self.grid, 1, 1)
        if isinstance(self.block, int):
            self.block = (self.block, 1, 1)



"""Define the kernel factory for the Conv2dGw8Wgrad kernel."""

from dataclasses import dataclass
from itertools import product
from typing import Generator

from .. import generators as gen
from ..util import divup, next_relative_prime
from ..cuda.driver import DeviceAttributes

from .launch_params import LaunchParams
from .conv2d_gw8_params import Conv2dGw8Params
from .conv2d_stats import Conv2dStats
from .kernel_factory import KernelFactory, KernelSpec
from .kernel import get_full_kernel_name


@dataclass(frozen=True)
class Conv2dGw8WgradConfig:
    """Tile configuration for the Conv2d GW8 Wgrad kernel."""

    groups: int = 8
    block_h: int = 16
    block_n_iters: int = 1
    warp_n: int = 1
    warp_s: int = None


KERNEL_NAME = "spio_conv2d_gw8_wgrad"

BLOCK_Q = 8


def _get_configs(
    params: Conv2dGw8Params,
    _device_attr: DeviceAttributes,
) -> Generator[Conv2dGw8WgradConfig, None, None]:
    """Generate tile configurations for the given layer parameters."""
    max_groups = min(params.groups, 8)
    max_warps = 32

    s_up = divup(params.s, 2) * 2
    block_w = BLOCK_Q + s_up - 1

    # # Try configurations with warp_s = params.S.
    block_h_values = [
        block_h for block_h in [2, 4, 8, 16, 32, 64] if block_h <= params.h
    ]
    if params.h not in block_h_values:
        block_h_values.append(params.h)
    groups_values = [groups for groups in [2, 4, 8] if groups <= max_groups]
    if params.groups not in groups_values and params.groups <= max_groups:
        groups_values.append(params.groups)
    warp_s_values = [warp_s for warp_s in [1, 2] if warp_s <= params.s]
    if params.s not in warp_s_values:
        warp_s_values.append(params.s)
    block_n_iters_values = [
        block_n_iters
        for block_n_iters in [1, 2, 4, 8, 16, 32]
        if block_n_iters <= params.n
    ]
    warp_n_values = [warp_n for warp_n in [1, 2, 4] if warp_n <= params.n]
    yield from (
        Conv2dGw8WgradConfig(
            groups=groups,
            block_h=block_h,
            warp_s=warp_s,
            block_n_iters=block_n_iters,
            warp_n=warp_n,
        )
        for groups, block_h, warp_s, block_n_iters, warp_n in product(
            groups_values,
            block_h_values,
            warp_s_values,
            block_n_iters_values,
            warp_n_values,
        )
        # Ensure that the number of groups does not exceed the hardware limit.
        if (groups * (divup(params.s, warp_s)) <= max_warps)
        # Ensure that a row of input values can be loaded with a single 128-bit load.
        and (warp_n * block_w <= 32 * divup(params.s, warp_s))
        # Avoid simulatenously large values of warp_n and groups.
        and (warp_n * groups <= max_groups * 2)
    )


def _get_kernel_spec(
    params: Conv2dGw8Params,
    config: Conv2dGw8WgradConfig,
    _device_attr: DeviceAttributes,
    **_kwargs,
) -> KernelSpec:
    """Get the code gen specs and launch params."""
    params.validate()

    r, s = params.r, params.s

    n, c, h, w = params.n, params.c, params.h, params.w
    p, q = params.p, params.q
    padding_h, padding_w = params.padding_h, params.padding_w
    transpose_padding_h = params.transpose_padding_h

    # Hardcoded parameters:
    group_width = params.group_width

    # Derived parameters
    c8 = c // 8
    groups = params.groups

    # Tiles
    block_n = config.block_n_iters * config.warp_n
    block_h = min(config.block_h, h)
    block_groups = min(config.groups, groups)

    # Derived Tiles
    s_up = divup(s, 2) * 2
    block_c = block_groups * group_width
    block_c8 = block_c // 8
    block_w = BLOCK_Q + s_up - 1
    block_p = block_h + r - 1
    blocks_n = divup(n, block_n)
    blocks_h = divup(h, block_h)
    blocks_q = divup(q, BLOCK_Q)
    blocks_c8 = divup(c8, block_c8)
    blocks = blocks_n * blocks_h * blocks_q * blocks_c8

    warps_c8 = block_c8
    warps_s = divup(s, config.warp_s)
    warps = warps_c8 * warps_s
    threads = warps * 32

    warp_s2 = config.warp_s // 2
    warp_s2_up = divup(config.warp_s, 2)

    # There are effectively 8 shared memory banks when storing 16-byte elements (8c float16).
    num_smem_banks = 8

    smem_x_stride = next_relative_prime(block_c8, num_smem_banks)

    smem_tensors = [
        gen.Tensor(
            "SmemInput",
            gen.dtype.uint4,
            {"ping_pong": 2, "n": config.warp_n, "x": block_w, "c8": block_c8},
            strides={"x": smem_x_stride},
        ),
        gen.Tensor(
            "SmemDelta",
            gen.dtype.uint4,
            {"ping_pong": 2, "n": config.warp_n, "q": BLOCK_Q, "k8": block_c8},
            strides={"x": smem_x_stride},
        ),
        gen.Tensor(
            "SmemWgrad", gen.dtype.float2, {"k8": block_c8, "s": s, "c": 8, "k2": 4}
        ),
    ]

    smem_size = max(
        smem_tensors[0].num_bytes + smem_tensors[1].num_bytes, smem_tensors[2].num_bytes
    )
    assert smem_size < _device_attr.max_shared_memory_per_block

    launch_params = LaunchParams(grid=blocks, block=threads)

    full_kernel_name = get_full_kernel_name(KERNEL_NAME, params)

    gen_specs = [
        gen.Macro({"SPIO_CONV_WGRAD_KERNEL": full_kernel_name}),
        #
        # Block parameters.
        #
        gen.Fold("block_n", "n", block_n),
        gen.Fold("block_y", "y", block_h),
        gen.Fold("block_p", "p", block_p),
        gen.Fold("block_q", "q", BLOCK_Q),
        gen.Fold("block_c", "c", block_c),
        gen.Fold("warp_s", "s", config.warp_s),
        gen.ParamsSpec(
            "Block",
            {
                "threads": threads,
            },
        ),
        #
        # Constant parameters.
        #
        gen.ParamsSpec(
            "Params",
            {
                "R_Size": r,
                "S_Size": s,
                "PADDING_H": padding_h,
                "PADDING_W": padding_w,
                "TRANSPOSE_PADDING_H": transpose_padding_h,
                "WARP_S_Size": config.warp_s,
                "WARP_S2_Size": warp_s2,
                "WARP_S2_UP_Size": warp_s2_up,
                "BLOCK_N_ITERS": config.block_n_iters,
                "WARP_N_Size": config.warp_n,
            },
        ),
        #
        # Block indices.
        #
        gen.Index(
            "BlockIdx",
            {
                "block_n": blocks_n,
                "block_y": blocks_h,
                "block_q": blocks_q,
                "block_c": blocks_c8,
            },
        ),
        #
        # Input loading.
        #
        gen.Index("InputIdx", {"n": config.warp_n, "x": block_w, "c8": block_c8}),
        gen.Tensor(
            "Input", gen.dtype.uint4, {"n": n, "y": h, "x": w, "c8": c8}, constant=True
        ),
        gen.Index(
            "SmemInputLoadIdx",
            {
                "c8": warps_c8,
                "warp_s": warps_s,
                "repeat": 32 // (2 * BLOCK_Q),
                "s": 2,
                "q": BLOCK_Q,
            },
            dummies=["repeat"],
        ),
        #
        # Delta loading
        #
        gen.Index("DeltaIdx", {"n": config.warp_n, "q": BLOCK_Q, "k8": block_c8}),
        gen.Tensor(
            "Delta", gen.dtype.uint4, {"n": n, "p": p, "q": q, "k8": c8}, constant=True
        ),
        gen.Index(
            "SmemDeltaLoadIdx",
            {
                "k8": warps_c8,
                "repeat": (32 * warps_s) // BLOCK_Q,
                "q": BLOCK_Q,
            },
            dummies=["repeat"],
        ),
        #
        # MMA fragments
        #
        gen.Fragment("Acc", gen.FragmentType.M16_N8_F32_C, "c", "k"),
        gen.Fragment("InputFragment", gen.FragmentType.M16_K8_F16_A, "c", "x"),
        gen.Fragment("DeltaFragment", gen.FragmentType.N8_K8_F16_B, "x", "k"),
        #
        # MMA tensors
        #
        gen.Tensor(
            "AccTensor", "Acc", {"s2": warp_s2_up, "r": r}
        ),
        gen.Tensor("InputTensor", "InputFragment", {"s2": warp_s2_up}),
        gen.Tensor("DeltaTensor", "DeltaFragment", {"n": config.warp_n, "r": r}),
        #
        # Weights storing.
        #
        gen.Index("SmemWgradStoreIdx", {"k8": warps_c8, "warp_s": warps_s, "lane": 32}),
        # Each thread stores 8k for a particular (k8, r, s, c).
        gen.Index("WgradStoreIdx", {"k8": warps_c8, "s": s, "c": 8}),
        # Reduce Wgrad through global memory using float32 precision.
        gen.Tensor("Wgrad", gen.dtype.float, {"k": c, "r": r, "s": s, "c": 8}),
    ] + smem_tensors
    return KernelSpec(gen_specs=gen_specs, launch_params=launch_params)


conv2d_gw8_wgrad_kernel_factory = KernelFactory(
    Conv2dGw8Params,
    Conv2dGw8WgradConfig,
    Conv2dStats,
    kernel_name=KERNEL_NAME,
    configs=_get_configs,
    kernel_spec=_get_kernel_spec,
    kernel_source_file="conv2d_gw8_wgrad.cu",
    src_module="spio.src",
    includes_module="spio.include",
    perf_model_skip_params=["group_width", "stride"],
)



"""Implement custom memory formats for PyTorch tensors."""

import torch

from typing import List


class SixteenChannelsLast:
    """Memory format for 16 channels last."""

    @staticmethod
    def format(tensor: torch.Tensor) -> torch.Tensor:
        """Transpose a tensor from [k, c] to [c16, k, 16c]."""
        if tensor.dim() == 2:
            return tensor.view(tensor.shape[0], -1, 16).permute(1, 0, 2).contiguous()
        elif tensor.dim() == 4:
            n, _, h, w = tensor.shape
            return tensor.view(n, -1, 16, h, w).permute(1, 0, 3, 4, 2).contiguous()
        else:
            raise NotImplementedError(f"{tensor.dim()}D tensor not supported")


def check_channels_last(args: List[torch.Tensor]):
    """Check that the arguments are channels_last tensors."""
    for arg in args:
        if isinstance(arg, torch.Tensor) and len(arg.shape) == 4:
            assert arg.is_contiguous(
                memory_format=torch.channels_last
            ), f"Tensor is not channels_last: {arg}"
